package com.amazon.inputsdk.demo;

import static android.graphics.Color.BLACK;
import static android.graphics.Color.RED;
import static android.graphics.Paint.Align.CENTER;
import static android.graphics.Paint.Align.LEFT;
import static android.view.KeyEvent.ACTION_DOWN;
import static android.view.KeyEvent.KEYCODE_DPAD_LEFT;
import static android.view.KeyEvent.KEYCODE_DPAD_RIGHT;
import static android.view.KeyEvent.KEYCODE_P;
import static android.view.KeyEvent.KEYCODE_S;
import static android.view.MotionEvent.ACTION_HOVER_MOVE;

import android.app.Activity;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Paint.Align;
import android.graphics.PointF;
import android.graphics.Rect;
import android.text.Layout;
import android.text.StaticLayout;
import android.text.TextPaint;
import android.util.Log;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.SurfaceHolder;
import android.view.SurfaceView;

import androidx.annotation.NonNull;

import com.amazon.device.inputmapping.help.TriggerHandler;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class Game extends SurfaceView implements SurfaceHolder.Callback {

    private GameThread gameThread;
    private List<Obstacle> obstacles;
    private Player player;
    private boolean isGameOver, isPaused, isMainScreenActive, isHoveredOnTooltip;
    private Context context;
    private Rect tooltipRect;

    private final List<Integer> obstaclePossibleXPos;

    private static final Random random = new Random(System.currentTimeMillis());
    private static final int MAX_OBSTACLES_NUMBER = 20;
    private static final float PLAYER_BASE_HEIGHT_PERCENTAGE = .85f;
    private static final float OBSTACLE_BASE_HEIGHT_PERCENTAGE = .10f;
    private static final float OBSTACLE_HEIGHT_BOUND_PERCENTAGE = .15f;
    private static final float TOOLTIP_MARGIN_PERCENTAGE = .98f;
    private static final int PLAYER_WIDTH = 75;
    private static final int PLAYER_HEIGHT = 100;
    private static final int OBSTACLE_RADIUS = 15;
    private static final int PLAYER_VELOCITY = 20;
    private static final int OBSTACLE_VELOCITY = 20;
    private static final float GAME_TEXT_SIZE = 50;
    private static final float TOOLTIP_TEXT_SIZE = 20;
    private static final int TEXT_BACKGROUND_MARGIN = 15;
    private static final int GAME_BACKGROUND = 0xFFDDDDDD;
    private static final int TOOLTIP_DIALOG_BACKGROUND = 0xFFFFE8AE;
    private static final int TOOLTIP_HEIGHT = 50;
    private static final int TOOLTIP_WIDTH = 50;
    private static final int TOOLTIP_DIALOG_WIDTH = 200;
    private static final String GAME_START = "Press S to Start";
    private static final String GAME_PAUSED = "Game Paused";
    private static final String GAME_OVER = "Game Over";
    private static final String TOOLTIP_HOVER_TEXT = "Press Shift + Tab to display the Input Map";


    public Game(Context context) {
        super(context);
        getHolder().addCallback(this);

        this.context = context;
        gameThread = new GameThread(getHolder(), this);
        isGameOver = false;
        isPaused = false;
        isMainScreenActive = true;
        isHoveredOnTooltip = false;

        obstaclePossibleXPos = new ArrayList<>();
        for(int i = OBSTACLE_RADIUS * 2; i <= getDisplayWidthPixels() - (OBSTACLE_RADIUS); i++) {
            obstaclePossibleXPos.add(i);
            i += (getDisplayWidthPixels() * .05);
        }

        obstacles = new ArrayList<>();
        for(int i = 0; i < MAX_OBSTACLES_NUMBER; i++) {
            addToList();
        }

        Bitmap playerSprite = BitmapFactory.decodeResource(getResources(), R.drawable.sprite);
        player = new Player
                (
                    new PointF(getScreenMidWidth(), getPlayerBaseHeight()),
                    PLAYER_VELOCITY, playerSprite, PLAYER_WIDTH, PLAYER_HEIGHT
                );

        setTooltipRect();

        setFocusable(true);
    }

    @Override
    public void surfaceCreated(@NonNull SurfaceHolder holder) {
        gameThread.setRunning(true);
        gameThread.start();
    }

    @Override
    public void surfaceChanged(@NonNull SurfaceHolder holder, int format, int width, int height) {

    }

    @Override
    public void surfaceDestroyed(@NonNull SurfaceHolder holder) {
        boolean retry = true;

        while(retry) {
            try {
                gameThread.setRunning(false);
                gameThread.join();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            retry = false;
        }
    }

    @Override
    public void draw(Canvas canvas) {
        invalidate();
        super.draw(canvas);
        canvas.drawColor(GAME_BACKGROUND);

        if(isGameOver) {
            drawText(canvas, GAME_OVER, getScreenMidWidth(), getScreenMidHeight(), BLACK, CENTER, GAME_TEXT_SIZE, GAME_BACKGROUND);
        }
        else if(isMainScreenActive) {
            drawText(canvas, GAME_START, getScreenMidWidth(), getScreenMidHeight(),  BLACK, CENTER, GAME_TEXT_SIZE, GAME_BACKGROUND);
        }
        else {
            drawObstacles(canvas, obstacles, RED);
            drawPlayer(canvas, player);
            if(isPaused) {
                drawText(canvas, GAME_PAUSED, getScreenMidWidth(), getScreenMidHeight(),  BLACK, CENTER, GAME_TEXT_SIZE, GAME_BACKGROUND);
            }
        }
        drawTooltip(canvas);
        if(isHoveredOnTooltip) {
            drawTooltipDialog(canvas);
        }
    }

    public void update() {
        if(!isPaused && !isMainScreenActive) {
            for (int i = obstacles.size(); i < MAX_OBSTACLES_NUMBER; i++) {
                addToList();
            }

            checkCollision();
            obstacles = getRemainingObstacles();
        }
    }

    @Override
    public boolean onKeyDown(int keyCode, KeyEvent event) {
        Log.i("Event", "Event: " + event);

        boolean isHandled = false;

        if(keyCode == KEYCODE_S) {
            isMainScreenActive = false;
        }
        if(keyCode == KEYCODE_P && !isMainScreenActive) {
            isPaused = !isPaused;
        }
        if(keyCode == KeyEvent.KEYCODE_TAB) {
            if (event.isShiftPressed()) {
                if (!isMainScreenActive && !isGameOver) {
                    isPaused = true;
                }
                TriggerHandler.getInstance().showHelp((Activity) context, "Test App");
                isHandled = true;
            }
        }

        if(!isPaused) {
            switch (keyCode) {
                case KEYCODE_DPAD_LEFT:
                    if(!isMainScreenActive) {
                        player.moveLeft(event);
                        isHandled = true;
                    }
                    break;
                case KEYCODE_DPAD_RIGHT:
                    if(!isMainScreenActive) {
                        player.moveRight(event);
                        isHandled = true;
                    }
                    break;
            }
        }

        return isHandled || super.onKeyUp(keyCode, event);
    }

    @Override
    public boolean onGenericMotionEvent(MotionEvent event) {
        boolean isHandled = false;
        if (event.getAction() == ACTION_HOVER_MOVE) {
            if (tooltipRect.contains((int) event.getX(), (int) event.getY())) {
                isHoveredOnTooltip = true;
                isHandled = true;
            } else {
                isHoveredOnTooltip = false;
            }
        }

        return isHandled || super.onGenericMotionEvent(event);
    }

    @Override
    public boolean onTouchEvent(MotionEvent event) {
        boolean isHandled = false;
        if(event.getAction() == ACTION_DOWN) {
            if (tooltipRect.contains((int) event.getX(), (int) event.getY())) {
                if (!isMainScreenActive && !isGameOver) {
                    isPaused = true;
                }
                TriggerHandler.getInstance().showHelp((Activity) context, "Test App");
                isHandled = true;
            }
        }

        return isHandled || super.onTouchEvent(event);
    }

    private int getDisplayWidthPixels() {
        return getResources().getDisplayMetrics().widthPixels;
    }

    private int getDisplayHeightPixels() {
        return getResources().getDisplayMetrics().heightPixels;
    }

    private void addToList() {
        float height = getObstaclesBaseHeight() +
                (
                        (-1 * getObstaclesHeightBound()) +
                        (2 * getObstaclesHeightBound()) * random.nextFloat()
                );
        PointF point = new PointF(obstaclePossibleXPos.get(random.nextInt(obstaclePossibleXPos.size())),
                height);
        obstacles.add(new Obstacle(point, OBSTACLE_RADIUS, OBSTACLE_VELOCITY));
    }

    private void checkCollision() {
        for(Obstacle obstacle: obstacles) {
            if(player.getRect().intersects(obstacle.getRect().left,
                    obstacle.getRect().top,
                    obstacle.getRect().right,
                    obstacle.getRect().bottom)) {
                isGameOver = true;
                return;
            }
        }
    }

    private List<Obstacle> getRemainingObstacles() {
        List<Obstacle> remainingObstacles = new ArrayList<>();
        for (Obstacle obstacle : obstacles) {
            if (obstacle.getPosition().y < getPlayerBaseHeight()) {
                remainingObstacles.add(obstacle);
                obstacle.move();
            }
        }
        return  remainingObstacles;
    }

    private void drawText(Canvas canvas, String text, int left, int bottom, int color, Align align, float textSize, int bgColor) {
        Paint textPaint = new Paint();
        textPaint.setTextSize(textSize);
        textPaint.setTextAlign(align);
        textPaint.setColor(color);

        Paint backgroundPaint = new Paint();
        backgroundPaint.setColor(bgColor);
        Rect bounds = new Rect();
        textPaint.getTextBounds(text, 0, text.length(), bounds);

        bounds.offsetTo(left - (bounds.width() / 2), bottom - bounds.height());
        bounds.set(
                bounds.left - TEXT_BACKGROUND_MARGIN,
                bounds.top - TEXT_BACKGROUND_MARGIN,
                bounds.right + TEXT_BACKGROUND_MARGIN,
                bounds.bottom + TEXT_BACKGROUND_MARGIN + 3
        );

        canvas.drawRect(bounds, backgroundPaint);
        canvas.drawText(text, left, bottom, textPaint);
    }

    private void drawObstacles(Canvas canvas, List<Obstacle> obstacles, int color) {
        Paint paint = new Paint();
        paint.setColor(color);

        for(Obstacle obstacle: obstacles) {
            canvas.drawCircle(obstacle.getPosition().x, obstacle.getPosition().y, obstacle.getRadius(), paint);
        }
    }

    private void drawPlayer(Canvas canvas, Player player) {
        canvas.drawBitmap(player.getImage(), null, player.getRect(), null);
    }

    private int getScreenMidWidth() {
        return getDisplayWidthPixels() / 2;
    }

    private int getScreenMidHeight() {
        return getDisplayHeightPixels() / 2;
    }

    private int getPlayerBaseHeight() {
        return (int) (getDisplayHeightPixels() * PLAYER_BASE_HEIGHT_PERCENTAGE);
    }

    private int getObstaclesBaseHeight() {
        return (int) (getDisplayHeightPixels() * OBSTACLE_BASE_HEIGHT_PERCENTAGE);
    }

    private int getObstaclesHeightBound() {
        return (int) (getDisplayHeightPixels() * OBSTACLE_HEIGHT_BOUND_PERCENTAGE);
    }

    private void setTooltipRect() {
        Rect tooltipRect = new Rect();
        int right = (int) (getDisplayWidthPixels() * TOOLTIP_MARGIN_PERCENTAGE);
        int bottom = (int) (getDisplayHeightPixels() * TOOLTIP_MARGIN_PERCENTAGE);
        tooltipRect.set(right - TOOLTIP_WIDTH, bottom - TOOLTIP_HEIGHT, right, bottom);

        this.tooltipRect = tooltipRect;
    }

    private void drawTooltip(Canvas canvas) {
        canvas.drawBitmap(BitmapFactory.decodeResource(getResources(), R.drawable.tooltip), null, tooltipRect, null);
    }

    private void drawTooltipDialog(Canvas canvas) {
        Paint backgroundPaint = new Paint();
        backgroundPaint.setColor(TOOLTIP_DIALOG_BACKGROUND);

        TextPaint textPaint = new TextPaint();
        textPaint.setTextSize(TOOLTIP_TEXT_SIZE);
        textPaint.setColor(BLACK);
        StaticLayout myStaticLayout = new StaticLayout(TOOLTIP_HOVER_TEXT, textPaint, TOOLTIP_DIALOG_WIDTH, Layout.Alignment.ALIGN_CENTER, 1, 0, false);

        canvas.save();
        canvas.translate(tooltipRect.left - myStaticLayout.getWidth(), tooltipRect.top - myStaticLayout.getHeight());
        Rect dialog = new Rect(0, 0, myStaticLayout.getWidth(), myStaticLayout.getHeight());
        canvas.drawRect(dialog, backgroundPaint);
        myStaticLayout.draw(canvas);
        canvas.restore();
    }
}
